import psycopg2
from psycopg2.extensions import AsIs


class TagExists(Exception):
    pass


class ODB(object):
    """class representing an Odoo instance
    """
    def __init__(self, db=None):
        self.db = db

    def connect(self, db=None):
        """ connect to the current db unless specified
        """
        if db is None:
            db = self.db
        return psycopg2.connect('dbname=%s' % db)

    def createdb(self):
        """ createdb used for tests
        """
        db = self.db
        with self.connect('postgres') as cn:
            cn.autocommit = True
            cn.cursor().execute('CREATE DATABASE "%s"', (AsIs(db),))
        with self.connect(db) as cn, cn.cursor() as cr:
            cr.execute("CREATE TABLE ir_config_parameter "
                       "(key character varying(256), value text)")

    def dropdb(self, db=None):
        """ drop db
        """
        if db is None:
            db = self.db
        with self.connect('postgres') as cn, cn.cursor() as cr:
            cn.autocommit = True
            self._disconnect(cr, db)
            cn.cursor().execute('DROP DATABASE "%s"', (AsIs(db),))

    def init(self):
        """ initialize the db with the revision
        """
        revision = self.get('revision')
        with self.connect() as cn, cn.cursor() as cr:
            if revision is None:
                revision = '1'
                self.set('revision', revision, cr)
                self.set('parent', '0', cr)
        return int(revision)

    def set(self, key, value, cr=None):
        """ set the value of a key
        """
        update = "UPDATE ir_config_parameter SET value=%s WHERE key=%s"
        insert = "INSERT INTO ir_config_parameter (value, key) values (%s, %s)"
        values = (str(value), 'odb.%s' % key)
        if cr is not None:
            cr.execute(update, values)
            if not cr.rowcount:
                cr.execute(insert, values)
        else:
            with self.connect() as cn, cn.cursor() as cr:
                cr.execute(update, values)
                if not cr.rowcount:
                    cr.execute(insert, values)

    def get(self, key, cr=None):
        """ get the value of a key
        """
        req = "SELECT value FROM ir_config_parameter WHERE key=%s"
        if cr is not None:
            cr.execute(req, ('odb.' + key,))
            res = cr.fetchone()
        else:
            with self.connect() as cn, cn.cursor() as cr:
                cr.execute(req, ('odb.' + key,))
                res = cr.fetchone()
        if res is not None and len(res) == 1:
            return res[0]

    def rem(self, key, cr=None):
        """ delete a key
        """
        req = "DELETE FROM ir_config_parameter WHERE key=%s"
        if cr is not None:
            cr.execute(req, ('odb.' + key,))
        else:
            with self.connect() as cn, cn.cursor() as cr:
                cr.execute(req, ('odb.' + key,))

    def revision(self):
        """ returns the db revision
        """
        return int(self.get('revision'))

    def parent(self):
        """ get the parent snapshot
        """
        return int(self.get('parent'))

    def _disconnect(self, cr, db):
        """ kill all pg connections
        """
        cr.execute("SELECT pg_terminate_backend(pg_stat_activity.pid) "
                   "FROM pg_stat_activity "
                   "WHERE pg_stat_activity.datname=%s "
                   "AND pid <> pg_backend_pid();", (db,))

    def commit(self):
        """ create a snapshot and change the current revision
        Corresponds to the commit command
        """
        revision = self.revision()
        targetdb = '*'.join([self.db, str(revision)])
        with self.connect('postgres') as cn, cn.cursor() as cr:
            cn.autocommit = True
            self._disconnect(cr, self.db)
            cr.execute('CREATE DATABASE "%s" WITH TEMPLATE "%s"', (AsIs(targetdb), AsIs(self.db)))
        self.set('revision', revision + 1)
        self.set('parent', revision)
        self.rem('tag')

    def revert(self, parent=None, tag=None):
        """ drop the current db and start back from this parent
        (or the current parent if no parent is specified)
        """
        if parent is None and tag is None:  # revert to last
            parent = self.parent()
        if tag:  # revert to tag
            tagfound = [r for r in self.log() if r.get('tag') == tag]
            if tagfound:
                parent = tagfound[0]['revision']
            else:
                return
        # store revision because we'll drop
        currevision = self.revision()
        sourcedb = '*'.join([self.db, str(parent)])
        with self.connect() as cn, cn.cursor() as cr:
            cn.autocommit = True
            self._disconnect(cr, self.db)
        with self.connect('postgres') as cn, cn.cursor() as cr:
            cn.autocommit = True
            cr.execute('DROP DATABASE "%s"', (AsIs(self.db),))
            cr.execute('CREATE DATABASE "%s" WITH TEMPLATE "%s"', (AsIs(self.db), AsIs(sourcedb)))
        self.set('revision', currevision)
        self.set('parent', parent)
        self.rem('tag')

    def log(self):
        """ return a list of previous revisions, each revision being a dict with needed infos
        """
        log = []
        with self.connect() as cn, cn.cursor() as cr:
            req = 'SELECT datname FROM pg_catalog.pg_database WHERE datname like %s'
            cr.execute(req, (self.db + '*%',))
            dbnames = cr.fetchall()
        for db in [d[0] for d in dbnames] + [self.db]:
            with self.connect(db) as cn, cn.cursor() as cr:
                log.append({
                    'db': db,
                    'revision': int(self.get('revision', cr)),
                    'parent': int(self.get('parent', cr)),
                })
                tag = self.get('tag', cr)
                if tag:
                    log[-1]['tag'] = tag
        return sorted(log, key=lambda x: x['revision'], reverse=True)

    def purge(self, what, confirm=False):
        """ purge the revisions
        what can be: 'all'
        """
        # first get what will be purged, then confirm
        to_purge = [l for l in self.log() if l['db'] != self.db]
        if what == 'all':
            pass
        else:
            raise NotImplementedError('Bad purge command')
        if confirm:
            for logitem in to_purge:
                self.dropdb(logitem['db'])
        return to_purge

    def tag(self, tag=None, revision=None, delete=False):
        """ tag a specific revision or the current one by default
        """
        tags = [r for r in self.log() if 'tag' in r]
        if delete:
            if tag in [r.get('tag') for r in tags]:
                db = [r['db'] for r in tags if r.get('tag') == tag][0]
                with self.connect(db) as cn, cn.cursor() as cr:
                    return self.rem('tag', cr)
            return
        if tag is None and revision is None:
            return tags
        if tag is not None and tag in [r.get('tag') for r in tags]:
            raise TagExists('This tag already exists')
        if revision is None:
            revision = self.revision()
        if self.revision() == revision:
            db = self.db
        else:
            db = '%s*%s' % (self.db, revision)
        with self.connect(db) as cn, cn.cursor() as cr:
            self.set('tag', tag, cr)



import os
import argparse
from .odb import ODB, TagExists
CONF = os.path.expanduser('~/.anybox.pg.odoo')


def main():
    parser = argparse.ArgumentParser(
        prog="odb",
        description=u"Postgresql snapshot versionning tool (for Odoo)",)
    subparsers = parser.add_subparsers(help='sub-commands')
    parser_init = subparsers.add_parser('init', help='Set the current db')
    parser_init.add_argument('db', metavar='db', nargs=1, help='database to work on')
    parser_commit = subparsers.add_parser('commit', help='Save the current db in a new revision')
    parser_info = subparsers.add_parser('info', help='Display the revision of the current db')
    parser_revert = subparsers.add_parser(
        'revert', help='Drop the current db and clone from a previous revision')
    parser_revert.add_argument('revision', nargs='?', help='revision to revert to')
    parser_log = subparsers.add_parser('log', help='List all available revisions')
    parser_purge = subparsers.add_parser('purge', help="Destroy revisions")
    parser_purge.add_argument('what', choices=['all'],
                              help='all: destroy all revisions except the current db')
    parser_purge.add_argument('-y', '--yes', action='store_true', help='Destroy without asking')
    parser_tags = subparsers.add_parser('tags', help="List all tags")
    parser_tag = subparsers.add_parser('tag', help="Tag a specific revision")
    parser_tag.add_argument('-d', '--delete', action='store_true', help='Delete tag')
    parser_tag.add_argument('tag', help='Tag')
    parser_tag.add_argument('revision', metavar='revision', nargs='?', help='Revision')

    def init(args):
        odb = ODB(args.db[0])
        odb.init()
        open(CONF, 'w').write(odb.db)
        print('Now revision %s' % odb.revision())

    def commit(args):
        odb = ODB(open(CONF).read())
        odb.commit()
        print('Now revision %s' % odb.revision())

    def revert(args):
        odb = ODB(open(CONF).read())
        if args.revision and args.revision.isdigit():
            odb.revert(parent=args.revision)
        elif args.revision and args.revision.isalnum():
            odb.revert(tag=args.revision)
        else:
            odb.revert()
        print('Reverted to parent %s, now at revision %s' % (odb.parent(), odb.revision()))

    def info(args):
        odb = ODB(open(CONF).read())
        print('database: %s' % odb.db)
        print('revision : %s (parent: %s)' % (odb.revision(), odb.parent()))

    def log(args):
        odb = ODB(open(CONF).read())
        for logitem in odb.log():
            print '%(db)s:\n\trevision: %(revision)s\n\tparent: %(parent)s' % logitem
            if 'tag' in logitem:
                print '\ttag: %s' % logitem['tag']

    def purge(args):
        odb = ODB(open(CONF).read())
        try:
            to_purge = odb.purge(args.what, args.yes)
        except NotImplementedError:
            print 'Unkown purge command'
            return
        print('Dropping these databases: %s' % ', '.join([i['db'] for i in to_purge]))
        if args.yes or raw_input('Confirm? y[N] ').lower() == 'y':
            odb.purge(args.what, True)
            print 'Purged'
        else:
            print 'Cancelled'

    def tags(args):
        odb = ODB(open(CONF).read())
        tags = odb.tag()
        for item in tags:
            print '%(tag)s (%(db)s)' % item

    def tag(args):
        odb = ODB(open(CONF).read())
        if args.delete:
            return odb.tag(args.tag, delete=True)
        try:
            odb.tag(args.tag, args.revision)
        except TagExists:
            print 'This tag already exists'

    parser_init.set_defaults(func=init)
    parser_commit.set_defaults(func=commit)
    parser_info.set_defaults(func=info)
    parser_revert.set_defaults(func=revert)
    parser_info.set_defaults(func=info)
    parser_log.set_defaults(func=log)
    parser_purge.set_defaults(func=purge)
    parser_tags.set_defaults(func=tags)
    parser_tag.set_defaults(func=tag)

    args = parser.parse_args()
    args.func(args)



import unittest
import time

from odb import ODB, TagExists


class TestCommit(unittest.TestCase):
    def setUp(self):
        """ create db
        """
        db = self.db = 'testodb-' + time.strftime('%Y%m%d%H%M%S')
        # create db
        ODB(db).createdb()

    def test_simple_commit(self):
        """ first simple scenario with commit and revert
        """
        odb = ODB(self.db)
        # init
        self.assertEqual(odb.init(), 1)
        self.assertEqual(odb.revision(), 1)
        self.assertEqual(odb.parent(), 0)
        # init
        self.assertEqual(odb.init(), 1)
        # commit
        odb.commit()
        self.assertEqual(odb.revision(), 2)
        self.assertEqual(odb.parent(), 1)
        # the db name remains the same
        self.assertEqual(self.db, odb.db)

        # commit
        odb.commit()
        self.assertEqual(odb.revision(), 3)
        self.assertEqual(odb.parent(), 2)

        # revert # TODO test the real revert
        odb.revert()
        self.assertEqual(odb.revision(), 3)
        self.assertEqual(odb.parent(), 2)

        # commit
        odb.commit()
        self.assertEqual(odb.revision(), 4)
        self.assertEqual(odb.parent(), 3)
        # just check that init again doesn't hurt
        self.assertEqual(odb.init(), 4)

        # revert to 2
        odb.revert(2)
        self.assertEqual(odb.revision(), 4)
        self.assertEqual(odb.parent(), 2)
        self.assertEqual(self.db, odb.db)

        # commit
        odb.commit()
        self.assertEqual(odb.revision(), 5)
        self.assertEqual(odb.parent(), 4)

        # list all revisions
        revs = odb.log()
        self.assertEqual(revs[-1]['revision'], 1)
        self.assertEqual(revs[1]['db'], self.db + '*4')
        self.assertEqual(revs[0]['revision'], 5)

        # tag revision 3 and 5
        odb.tag('v1', 3)
        odb.tag('v2')
        self.assertRaises(TagExists, odb.tag, 'v1', 4)

        # tags appear in the log
        self.assertEqual(odb.log()[2]['tag'], 'v1')
        self.assertEqual(odb.log()[0]['tag'], 'v2')

        # delete tag
        odb.tag('v2', delete=True)
        self.assertEqual(odb.log()[0].get('tag'), None)

        # we can revert to a tag
        odb.revert(tag='v1')
        revs = odb.log()
        self.assertEqual(revs[0]['revision'], 5)
        self.assertEqual(revs[0]['parent'], 3)
        # the current db has not the same tag
        self.assertEqual(odb.get('tag'), None)
        # even after commit, no tag
        odb.tag('v3')
        odb.commit()
        self.assertEqual(odb.get('tag'), None)

        # purge without confirmation does nothing
        odb.purge('all')
        self.assertEqual(len(odb.log()), 6)

        # check for real
        odb.purge('all', confirm=True)
        self.assertEqual(len(odb.log()), 1)

    def tearDown(self):
        """ cleanup
        """
        # drop everything (even if tests failed)
        odb = ODB(self.db)
        odb.purge('all', confirm=True)
        odb.dropdb()
